@@ -1,33 +1,42 @@
 /*
- * Copyright 2005-2018 Gentoo Foundation
+ * Copyright 2005-2019 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
  * Copyright 2005-2014 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2017-     Fabian Groffen  - <grobian@gentoo.org>
  */
 
-#if defined(__MACH__)
+#include "main.h"
+#include "applets.h"
+
+#include <unistd.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
+
+#if defined(__MACH__) && defined(__APPLE__)
 #include <libproc.h>
 #endif
 
-#define Q_FLAGS "irmM:" COMMON_FLAGS
+#include "basename.h"
+#include "eat_file.h"
+#include "rmspace.h"
+
+#define Q_FLAGS "io" COMMON_FLAGS
 static struct option const q_long_opts[] = {
 	{"install",       no_argument, NULL, 'i'},
-	{"reinitialize", opt_argument, NULL, 'r'},
-	{"metacache",    opt_argument, NULL, 'm'},
-	{"modpath",        a_argument, NULL, 'M'},
+	{"overlays",      no_argument, NULL, 'o'},
 	COMMON_LONG_OPTS
 };
 static const char * const q_opts_help[] = {
 	"Install symlinks for applets",
-	"Reinitialize ebuild cache",
-	"Reinitialize metadata cache",
-	"Module path",
+	"Print available overlays (read from repos.conf)",
 	COMMON_OPTS_HELP
 };
 #define q_usage(ret) usage(ret, Q_FLAGS, q_long_opts, q_opts_help, NULL, lookup_applet_idx("q"))
 
-static APPLET lookup_applet(const char *applet)
+APPLET lookup_applet(const char *applet)
 {
 	unsigned int i;
 
@@ -37,7 +46,8 @@ static APPLET lookup_applet(const char *applet)
 	for (i = 0; applets[i].name; ++i) {
 		if (strcmp(applets[i].name, applet) == 0) {
 			argv0 = applets[i].name;
-			if (i && applets[i].desc != NULL) ++argv0; /* chop the leading 'q' */
+			if (i && applets[i].desc != NULL)
+				++argv0; /* chop the leading 'q' */
 			return applets[i].func;
 		}
 	}
@@ -65,7 +75,9 @@ int lookup_applet_idx(const char *applet)
 
 int q_main(int argc, char **argv)
 {
-	int i, install;
+	int i;
+	bool install;
+	bool print_overlays;
 	const char *p;
 	APPLET func;
 
@@ -82,31 +94,13 @@ int q_main(int argc, char **argv)
 	if (argc == 1)
 		q_usage(EXIT_FAILURE);
 
-	install = 0;
-
+	install = false;
+	print_overlays = false;
 	while ((i = GETOPT_LONG(Q, q, "+")) != -1) {
 		switch (i) {
 		COMMON_GETOPTS_CASES(q)
-		case 'M': modpath = optarg; break;
-		case 'm':
-			if (optarg) {
-				const char *path = initialize_flat(optarg, CACHE_METADATA, true);
-				if (USE_CLEANUP)
-					free((void *)path);
-				reinitialize_metacache = -1;
-			} else
-				reinitialize_metacache = 1;
-			break;
-		case 'r':
-			if (optarg) {
-				const char *path = initialize_flat(optarg, CACHE_EBUILD, true);
-				if (USE_CLEANUP)
-					free((void *)path);
-				reinitialize = -1;
-			} else
-				reinitialize = 1;
-			break;
-		case 'i': install = 1; break;
+		case 'i': install = true;        break;
+		case 'o': print_overlays = true; break;
 		}
 	}
 
@@ -119,7 +113,7 @@ int q_main(int argc, char **argv)
 		if (!quiet)
 			printf("Installing symlinks:\n");
 
-#if defined(__MACH__)
+#if defined(__MACH__) && defined(__APPLE__)
 		rret = proc_pidpath(getpid(), buf, sizeof(buf));
 		if (rret != -1)
 			rret = strlen(buf);
@@ -129,13 +123,13 @@ int q_main(int argc, char **argv)
 		if ((size_t)rret > sizeof(buf) - 1) {
 			rret = -1;
 		} else {
-			strncpy(buf, prog, rret);
+			snprintf(buf, sizeof(buf), "%s", prog);
 		}
 #else
 		rret = readlink("/proc/self/exe", buf, sizeof(buf) - 1);
 #endif
 		if (rret == -1) {
-			warnfp("haha no symlink love for you");
+			warnfp("haha no symlink love for you ... :(");
 			return 1;
 		}
 		buf[rret] = '\0';
@@ -152,7 +146,8 @@ int q_main(int argc, char **argv)
 		for (i = 1; applets[i].desc; ++i) {
 			int r = symlinkat(prog, fd, applets[i].name);
 			if (!quiet)
-				printf(" %s ...\t[%s]\n", applets[i].name, r ? strerror(errno) : "OK");
+				printf(" %s ...\t[%s]\n",
+						applets[i].name, r ? strerror(errno) : "OK");
 			if (r && errno != EEXIST)
 				ret = 1;
 		}
@@ -162,12 +157,36 @@ int q_main(int argc, char **argv)
 		return ret;
 	}
 
-	if (reinitialize > 0 || reinitialize_metacache > 0)
-		return 0;
-	if (reinitialize < 0 || reinitialize_metacache < 0) {
-		reinitialize = reinitialize_metacache = 0;
+	if (print_overlays) {
+		char *overlay;
+		char *repo_name = NULL;
+		size_t repo_name_len = 0;
+		char buf[_Q_PATH_MAX];
+		size_t n;
+
+		array_for_each(overlays, n, overlay) {
+			repo_name = xarrayget(overlay_names, n);
+			if (strcmp(repo_name, "<PORTDIR>") == 0) {
+				repo_name = NULL;
+				snprintf(buf, sizeof(buf), "%s/profiles/repo_name", overlay);
+				if (!eat_file(buf, &repo_name, &repo_name_len))
+					repo_name = NULL;
+				if (repo_name != NULL)
+					rmspace(repo_name);
+			}
+			printf("%s%s%s: %s%s%s%s\n",
+					GREEN, repo_name == NULL ? "?unknown?" : repo_name,
+					NORM, overlay,
+					YELLOW, main_overlay == overlay ? " (main)" : "", NORM);
+			if (repo_name_len != 0) {
+				free(repo_name);
+				repo_name_len = 0;
+			}
+		}
+
 		return 0;
 	}
+
 	if (argc == optind)
 		q_usage(EXIT_FAILURE);
 	if ((func = lookup_applet(argv[optind])) == NULL)
@@ -186,43 +205,3 @@ int q_main(int argc, char **argv)
 
 	return (func)(argc - 1, argv);
 }
-
-static int run_applet_l(const char *arg, ...)
-{
-	int (*applet)(int, char **);
-	va_list ap;
-	int ret, optind_saved, argc;
-	char **argv;
-	const char *argv0_saved;
-
-	optind_saved = optind;
-	argv0_saved = argv0;
-
-	applet = lookup_applet(arg);
-	if (!applet)
-		return -1;
-
-	/* This doesn't NULL terminate argv, but you should be using argc */
-	va_start(ap, arg);
-	argc = 0;
-	argv = NULL;
-	while (arg) {
-		argv = xrealloc(argv, sizeof(*argv) * ++argc);
-		argv[argc - 1] = xstrdup(arg);
-		arg = va_arg(ap, const char *);
-	}
-	va_end(ap);
-
-	optind = 0;
-	argv0 = argv[0];
-	ret = applet(argc, argv);
-
-	while (argc--)
-		free(argv[argc]);
-	free(argv);
-
-	optind = optind_saved;
-	argv0 = argv0_saved;
-
-	return ret;
-}
