@@ -19,12 +19,13 @@
 
 #include "atom.h"
 #include "eat_file.h"
+#include "set.h"
 #include "xarray.h"
 #include "xasprintf.h"
 
 #define QLOP_DEFAULT_LOGFILE "emerge.log"
 
-#define QLOP_FLAGS "ctaHMmuUslerd:f:w:F:" COMMON_FLAGS
+#define QLOP_FLAGS "ctaHMmuUsElerd:f:w:F:" COMMON_FLAGS
 static struct option const qlop_long_opts[] = {
 	{"summary",   no_argument, NULL, 'c'},
 	{"time",      no_argument, NULL, 't'},
@@ -35,6 +36,7 @@ static struct option const qlop_long_opts[] = {
 	{"unmerge",   no_argument, NULL, 'u'},
 	{"autoclean", no_argument, NULL, 'U'},
 	{"sync",      no_argument, NULL, 's'},
+	{"emerge",    no_argument, NULL, 'E'},
 	{"endtime",   no_argument, NULL, 'e'},
 	{"running",   no_argument, NULL, 'r'},
 	{"date",       a_argument, NULL, 'd'},
@@ -54,6 +56,7 @@ static const char * const qlop_opts_help[] = {
 	"Show unmerge history",
 	"Show autoclean unmerge history",
 	"Show sync history",
+	"Show last merge like how emerge(1) -v would show it",
 	"Report time at which the operation finished (iso started)",
 	"Show current emerging packages",
 	"Limit selection to this time (1st -d is start, 2nd -d is end)",
@@ -85,6 +88,7 @@ struct qlop_mode {
 	char do_machine:1;
 	char do_endtime:1;
 	char show_lastmerge:1;
+	char show_emerge:1;
 	const char *fmt;
 };
 
@@ -275,6 +279,25 @@ static char *fmt_elapsedtime(struct qlop_mode *flags, time_t e)
 	return _elapsed_buf;
 }
 
+struct pkg_match {
+	char id[BUFSIZ];
+	depend_atom *atom;
+	time_t tbegin;
+	time_t time;
+	size_t cnt;
+};
+
+static int
+pkg_sort_cb(const void *l, const void *r)
+{
+	struct pkg_match *pl = *(struct pkg_match **)l;
+	struct pkg_match *pr = *(struct pkg_match **)r;
+	depend_atom *al = pl->atom;
+	depend_atom *ar = pr->atom;
+
+	return atom_compar_cb(al, ar);
+}
+
 /* The format of the sync log has changed over time.
 
 Old format:
@@ -326,6 +349,7 @@ static int do_emerge_log(
 	char *p;
 	char *q;
 	time_t tstart = LONG_MAX;
+	time_t tstart_emerge = 0;
 	time_t last_merge = 0;
 	time_t sync_start = 0;
 	time_t sync_time = 0;
@@ -333,31 +357,35 @@ static int do_emerge_log(
 	time_t elapsed;
 	depend_atom *atom;
 	depend_atom *atomw;
+	depend_atom *upgrade_atom = NULL;
 	DECLARE_ARRAY(merge_matches);
-	DECLARE_ARRAY(merge_averages);
+	set *merge_averages = create_set();
 	DECLARE_ARRAY(unmerge_matches);
-	DECLARE_ARRAY(unmerge_averages);
+	set *unmerge_averages = create_set();
+	set *atomset = NULL;
 	size_t i;
 	size_t parallel_emerge = 0;
 	bool all_atoms = false;
-
-	struct pkg_match {
-		char id[BUFSIZ];
-		depend_atom *atom;
-		time_t tbegin;
-		time_t time;
-		size_t cnt;
-	};
+	char afmt[BUFSIZ];
 	struct pkg_match *pkg;
 	struct pkg_match *pkgw;
 
-	if ((fp = fopen(log, "r")) == NULL) {
+	/* support relative path in here and now, when using ROOT, stick to
+	 * it, turning relative into a moot point */
+	if (portroot[1] == '\0')
+		snprintf(buf, sizeof(buf), "%s", log);
+	else
+		snprintf(buf, sizeof(buf), "%s%s", portroot, log);
+	if ((fp = fopen(buf, "r")) == NULL)
+	{
 		warnp("Could not open logfile '%s'", log);
 		return 1;
 	}
 
 	all_atoms = array_cnt(atoms) == 0;
 	if (all_atoms || flags->show_lastmerge) {
+		atomset = create_set();
+
 		/* assemble list of atoms */
 		while (fgets(buf, sizeof(buf), fp) != NULL) {
 			if ((p = strchr(buf, ':')) == NULL)
@@ -369,12 +397,8 @@ static int do_emerge_log(
 				continue;
 
 			if (flags->show_lastmerge) {
-				if (strncmp(p, "  *** emerge ", 13) == 0) {
-					last_merge = tstart;
-					array_for_each(atoms, i, atomw)
-						atom_implode(atomw);
-					xarrayfree_int(atoms);
-				}
+				if (strncmp(p, "  *** emerge ", 13) == 0)
+					tstart_emerge = tstart;
 				if (!all_atoms)
 					continue;
 			}
@@ -409,18 +433,26 @@ static int do_emerge_log(
 				atom->PV = NULL;
 				atom->PVR = NULL;
 				atom->PR_int = 0;
+				snprintf(afmt, sizeof(afmt), "%s/%s", atom->CATEGORY, atom->PN);
 
-				atomw = NULL;
-				array_for_each(atoms, i, atomw) {
-					if (atom_compare(atom, atomw) == EQUAL)
-						break;
-					atomw = NULL;
+				/* now we found a package, register this merge as a
+				 * "valid" one, such that dummy emerge calls (e.g.
+				 * emerge -pv foo) are ignored */
+				if (last_merge != tstart_emerge) {
+					array_t vals;
+
+					values_set(atomset, &vals);
+					array_for_each(&vals, i, atomw)
+						atom_implode(atomw);
+					xarrayfree_int(&vals);
+
+					clear_set(atomset);
+					last_merge = tstart_emerge;
 				}
-				if (atomw == NULL) {
-					xarraypush_ptr(atoms, atom);
-				} else {
+
+				atomw = add_set_value(afmt, atom, atomset);
+				if (atomw != NULL)
 					atom_implode(atom);
-				}
 			}
 		}
 
@@ -442,8 +474,7 @@ static int do_emerge_log(
 		 * finished without knowledge of another instance */
 		if (flags->do_running &&
 				(strncmp(p, "  *** emerge ", 13) == 0 ||
-				 strncmp(p, "  *** terminating.", 18) == 0 ||
-				 strncmp(p, "  *** exiting ", 14) == 0))
+				 strncmp(p, "  *** terminating.", 18) == 0))
 		{
 			if (p[7] == 'm') {
 				parallel_emerge++;
@@ -558,10 +589,16 @@ static int do_emerge_log(
 
 				/* see if we need this atom */
 				atomw = NULL;
-				array_for_each(atoms, i, atomw) {
-					if (atom_compare(atom, atomw) == EQUAL)
-						break;
-					atomw = NULL;
+				if (atomset == NULL) {
+					array_for_each(atoms, i, atomw) {
+						if (atom_compare(atom, atomw) == EQUAL)
+							break;
+						atomw = NULL;
+					}
+				} else {
+					snprintf(afmt, sizeof(afmt), "%s/%s",
+							atom->CATEGORY, atom->PN);
+					atomw = get_set(afmt, atomset);
 				}
 				if (atomw == NULL) {
 					atom_implode(atom);
@@ -586,34 +623,31 @@ static int do_emerge_log(
 					if (flags->do_average || flags->do_running)
 					{
 						/* find in list of averages */
-						size_t n;
-
-						pkg = NULL;
-						array_for_each(merge_averages, n, pkg) {
-							if (atom_compare(pkg->atom, pkgw->atom) == EQUAL) {
-								pkg->cnt++;
-								pkg->time += elapsed;
-								/* store max time for do_running */
-								if (elapsed > pkg->tbegin)
-									pkg->tbegin = elapsed;
-								atom_implode(pkgw->atom);
-								xarraydelete(merge_matches, i);
-								break;
-							}
-							pkg = NULL;
+						if (!verbose || flags->do_running) {
+							snprintf(afmt, sizeof(afmt), "%s/%s",
+								pkgw->atom->CATEGORY, pkgw->atom->PN);
+						} else {
+							snprintf(afmt, sizeof(afmt), "%s/%s-%s",
+								pkgw->atom->CATEGORY, pkgw->atom->PN,
+								pkgw->atom->PR_int > 0 ?
+								pkgw->atom->PVR : pkgw->atom->PV);
 						}
-						if (pkg == NULL) {  /* push new entry */
-							if (!verbose || flags->do_running) {
-								/* strip off version info */
-								pkgw->atom->PV = NULL;
-								pkgw->atom->PVR = NULL;
-								pkgw->atom->PR_int = 0;
-							}
+
+						pkg = add_set_value(afmt, pkgw, merge_averages);
+						if (pkg != NULL) {
+							pkg->cnt++;
+							pkg->time += elapsed;
+							/* store max time for do_running */
+							if (elapsed > pkg->tbegin)
+								pkg->tbegin = elapsed;
+							atom_implode(pkgw->atom);
+							xarraydelete(merge_matches, i);
+						} else {
+							/* new entry */
 							pkgw->id[0] = '\0';
 							pkgw->cnt = 1;
 							pkgw->time = elapsed;
 							pkgw->tbegin = elapsed;
-							xarraypush_ptr(merge_averages, pkgw);
 							xarraydelete_ptr(merge_matches, i);
 						}
 						break;
@@ -624,6 +658,34 @@ static int do_emerge_log(
 								flags->do_time ? ": " : "",
 								flags->do_time ?
 									fmt_elapsedtime(flags, elapsed) : "");
+					} else if (flags->show_emerge) {
+						int state = NOT_EQUAL;
+						if (upgrade_atom != NULL)
+							state = atom_compare(pkgw->atom, upgrade_atom);
+						switch (state) {
+							/*         "NRUD " */
+							case EQUAL:
+								printf(" %sR%s   ", YELLOW, NORM);
+								break;
+							case NOT_EQUAL:
+								printf("%sN%s    ", GREEN, NORM);
+								break;
+							case NEWER:
+								printf("  %sU%s  ", BLUE, NORM);
+								break;
+							case OLDER:
+								printf("  %sU%sD%s ", BLUE, DKBLUE, NORM);
+								break;
+						}
+						printf("%s", atom_format(flags->fmt, pkgw->atom));
+						if (state == NEWER || state == OLDER)
+							printf(" %s[%s%s%s]%s", DKBLUE, NORM,
+									atom_format("%[PVR]", upgrade_atom),
+									DKBLUE, NORM);
+						if (flags->do_time)
+							printf(": %s\n", fmt_elapsedtime(flags, elapsed));
+						else
+							printf("\n");
 					} else if (flags->do_time) {
 						printf("%s >>> %s: %s\n",
 								fmt_date(flags, pkgw->tbegin, tstart),
@@ -661,12 +723,26 @@ static int do_emerge_log(
 				if (atom == NULL)
 					continue;
 
+				if (p[-20] == ' ' && flags->show_emerge)
+				{
+					if (upgrade_atom != NULL)
+						atom_implode(upgrade_atom);
+					upgrade_atom = atom;
+					continue;
+				}
+
 				/* see if we need this atom */
 				atomw = NULL;
-				array_for_each(atoms, i, atomw) {
-					if (atom_compare(atom, atomw) == EQUAL)
-						break;
-					atomw = NULL;
+				if (atomset == NULL) {
+					array_for_each(atoms, i, atomw) {
+						if (atom_compare(atom, atomw) == EQUAL)
+							break;
+						atomw = NULL;
+					}
+				} else {
+					snprintf(afmt, sizeof(afmt), "%s/%s",
+							atom->CATEGORY, atom->PN);
+					atomw = get_set(afmt, atomset);
 				}
 				if (atomw == NULL) {
 					atom_implode(atom);
@@ -691,34 +767,31 @@ static int do_emerge_log(
 					if (flags->do_average || flags->do_running)
 					{
 						/* find in list of averages */
-						size_t n;
-
-						pkg = NULL;
-						array_for_each(unmerge_averages, n, pkg) {
-							if (atom_compare(pkg->atom, pkgw->atom) == EQUAL) {
-								pkg->cnt++;
-								pkg->time += elapsed;
-								/* store max time for do_running */
-								if (elapsed > pkg->tbegin)
-									pkg->tbegin = elapsed;
-								atom_implode(pkgw->atom);
-								xarraydelete(unmerge_matches, i);
-								break;
-							}
-							pkg = NULL;
+						if (!verbose || flags->do_running) {
+							snprintf(afmt, sizeof(afmt), "%s/%s",
+								pkgw->atom->CATEGORY, pkgw->atom->PN);
+						} else {
+							snprintf(afmt, sizeof(afmt), "%s/%s-%s",
+								pkgw->atom->CATEGORY, pkgw->atom->PN,
+								pkgw->atom->PR_int > 0 ?
+								pkgw->atom->PVR : pkgw->atom->PV);
 						}
-						if (pkg == NULL) {  /* push new entry */
-							if (!verbose || flags->do_running) {
-								/* strip off version info */
-								pkgw->atom->PV = NULL;
-								pkgw->atom->PVR = NULL;
-								pkgw->atom->PR_int = 0;
-							}
+
+						pkg = add_set_value(afmt, pkgw, unmerge_averages);
+						if (pkg != NULL) {
+							pkg->cnt++;
+							pkg->time += elapsed;
+							/* store max time for do_running */
+							if (elapsed > pkg->tbegin)
+								pkg->tbegin = elapsed;
+							atom_implode(pkgw->atom);
+							xarraydelete(unmerge_matches, i);
+						} else {
+							/* new entry */
 							pkgw->id[0] = '\0';
 							pkgw->cnt = 1;
 							pkgw->time = elapsed;
 							pkgw->tbegin = elapsed;
-							xarraypush_ptr(unmerge_averages, pkgw);
 							xarraydelete_ptr(unmerge_matches, i);
 						}
 						break;
@@ -771,22 +844,20 @@ static int do_emerge_log(
 			}
 		}
 		array_for_each(merge_matches, i, pkgw) {
-			size_t j;
 			time_t maxtime = 0;
 			bool isMax = false;
 
+			snprintf(afmt, sizeof(afmt), "%s/%s",
+					pkgw->atom->CATEGORY, pkgw->atom->PN);
+
 			elapsed = tstart - pkgw->tbegin;
-			pkg = NULL;
-			array_for_each(merge_averages, j, pkg) {
-				if (atom_compare(pkg->atom, pkgw->atom) == EQUAL) {
-					maxtime = pkg->time / pkg->cnt;
-					if (elapsed >= maxtime) {
-						maxtime = elapsed >= pkg->tbegin ? 0 : pkg->tbegin;
-						isMax = true;
-					}
-					break;
+			pkg = get_set(afmt, merge_averages);
+			if (pkg != NULL) {
+				maxtime = pkg->time / pkg->cnt;
+				if (elapsed >= maxtime) {
+					maxtime = elapsed >= pkg->tbegin ? 0 : pkg->tbegin;
+					isMax = true;
 				}
-				pkg = NULL;
 			}
 
 			/* extract (X of Y) from id, bug #442406 */
@@ -817,22 +888,20 @@ static int do_emerge_log(
 						isMax ? " (longest run)" : " (average run)" : "");
 		}
 		array_for_each(unmerge_matches, i, pkgw) {
-			size_t j;
 			time_t maxtime = 0;
 			bool isMax = false;
 
+			snprintf(afmt, sizeof(afmt), "%s/%s",
+					pkgw->atom->CATEGORY, pkgw->atom->PN);
+
 			elapsed = tstart - pkgw->tbegin;
-			pkg = NULL;
-			array_for_each(unmerge_averages, j, pkg) {
-				if (atom_compare(pkg->atom, pkgw->atom) == EQUAL) {
-					maxtime = pkg->time / pkg->cnt;
-					if (elapsed >= maxtime) {
-						maxtime = elapsed >= pkg->tbegin ? 0 : pkg->tbegin;
-						isMax = true;
-					}
-					break;
+			pkg = get_set(afmt, unmerge_averages);
+			if (pkg != NULL) {
+				maxtime = pkg->time / pkg->cnt;
+				if (elapsed >= maxtime) {
+					maxtime = elapsed >= pkg->tbegin ? 0 : pkg->tbegin;
+					isMax = true;
 				}
-				pkg = NULL;
 			}
 
 			if (flags->do_time) {
@@ -855,8 +924,11 @@ static int do_emerge_log(
 		size_t total_merges = 0;
 		size_t total_unmerges = 0;
 		time_t total_time = (time_t)0;
+		DECLARE_ARRAY(avgs);
 
-		array_for_each(merge_averages, i, pkg) {
+		values_set(merge_averages, avgs);
+		xarraysort(avgs, pkg_sort_cb);
+		array_for_each(avgs, i, pkg) {
 			printf("%s: %s average for %s%zd%s merge%s\n",
 					atom_format(flags->fmt, pkg->atom),
 					fmt_elapsedtime(flags, pkg->time / pkg->cnt),
@@ -864,7 +936,11 @@ static int do_emerge_log(
 			total_merges += pkg->cnt;
 			total_time += pkg->time;
 		}
-		array_for_each(unmerge_averages, i, pkg) {
+		xarrayfree_int(avgs);
+
+		values_set(unmerge_averages, avgs);
+		xarraysort(avgs, pkg_sort_cb);
+		array_for_each(avgs, i, pkg) {
 			printf("%s: %s average for %s%zd%s unmerge%s\n",
 					atom_format(flags->fmt, pkg->atom),
 					fmt_elapsedtime(flags, pkg->time / pkg->cnt),
@@ -872,6 +948,8 @@ static int do_emerge_log(
 			total_unmerges += pkg->cnt;
 			total_time += pkg->time;
 		}
+		xarrayfree_int(avgs);
+
 		if (sync_cnt > 0) {
 			printf("%ssync%s: %s average for %s%zd%s sync%s\n",
 					BLUE, NORM, fmt_elapsedtime(flags, sync_time / sync_cnt),
@@ -928,6 +1006,7 @@ int qlop_main(int argc, char **argv)
 	m.do_machine = 0;
 	m.do_endtime = 0;
 	m.show_lastmerge = 0;
+	m.show_emerge = 0;
 	m.fmt = NULL;
 
 	while ((ret = GETOPT_LONG(QLOP, qlop, "")) != -1) {
@@ -939,6 +1018,12 @@ int qlop_main(int argc, char **argv)
 			case 'u': m.do_unmerge = 1;     break;
 			case 'U': m.do_autoclean = 1;   break;
 			case 's': m.do_sync = 1;        break;
+			case 'E': m.do_merge = 1;
+					  m.do_unmerge = 1;
+					  m.do_autoclean = 1;
+					  m.show_lastmerge = 1;
+					  m.show_emerge = 1;
+					  verbose = 1;          break;
 			case 'r': m.do_running = 1;     break;
 			case 'a': m.do_average = 1;     break;
 			case 'c': m.do_summary = 1;     break;
