@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2018 Gentoo Authors
+ * Copyright 2005-2019 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -7,17 +7,34 @@
  * Copyright 2018-     Fabian Groffen  - <grobian@gentoo.org>
  */
 
-#ifdef APPLET_qmerge
+#include "main.h"
+#include "applets.h"
 
-/* This is a GNUlib hack, because GNUlib doesn't provide st_mtim members
- * of struct stat, but instead provides wrappers to retrieve the time
- * fields (stat-time module). We just define a macro in case people are
- * building without autoconf. */
-#ifdef _GL_SYS_STAT_H
-# include "stat-time.h"
-#else
-# define get_stat_mtime(X) ((X)->st_mtim)
-#endif
+#include <stdio.h>
+#include <xalloc.h>
+#include <fnmatch.h>
+#include <dirent.h>
+#include <string.h>
+#include <sys/types.h>
+#include <sys/wait.h>
+#include <assert.h>
+#include "stat-time.h"
+
+#include "atom.h"
+#include "copy_file.h"
+#include "contents.h"
+#include "eat_file.h"
+#include "human_readable.h"
+#include "md5_sha1_sum.h"
+#include "profile.h"
+#include "rmspace.h"
+#include "scandirat.h"
+#include "set.h"
+#include "tree.h"
+#include "xasprintf.h"
+#include "xchdir.h"
+#include "xmkdir.h"
+#include "xsystem.h"
 
 #ifndef GLOB_BRACE
 # define GLOB_BRACE     (1 << 10)	/* Expand "{a,b}" to "a" "b".  */
@@ -101,10 +118,65 @@ typedef struct llist_char_t llist_char;
 
 static void pkg_fetch(int, const depend_atom *, const struct pkg_t *);
 static void pkg_merge(int, const depend_atom *, const struct pkg_t *);
-static int pkg_unmerge(q_vdb_pkg_ctx *, queue *, int, char **, int, char **);
+static int pkg_unmerge(tree_pkg_ctx *, set *, int, char **, int, char **);
 static struct pkg_t *grab_binpkg_info(const char *);
 static char *find_binpkg(const char *);
 
+static bool
+prompt(const char *p)
+{
+	printf("%s? [Y/n] ", p);
+	fflush(stdout);
+	switch (getc(stdin)) {
+	case '\n':
+	case 'y':
+	case 'Y':
+		return true;
+	default:
+		return false;
+	}
+}
+
+static int run_applet_l(const char *arg, ...)
+{
+	int (*applet)(int, char **);
+	va_list ap;
+	int ret, optind_saved, argc;
+	char **argv;
+	const char *argv0_saved;
+
+	optind_saved = optind;
+	argv0_saved = argv0;
+
+	applet = lookup_applet(arg);
+	if (!applet)
+		return -1;
+
+	/* This doesn't NULL terminate argv, but you should be using argc */
+	va_start(ap, arg);
+	argc = 0;
+	argv = NULL;
+	while (arg) {
+		argv = xrealloc(argv, sizeof(*argv) * ++argc);
+		argv[argc - 1] = xstrdup(arg);
+		arg = va_arg(ap, const char *);
+	}
+	va_end(ap);
+
+	optind = 0;
+	argv0 = argv[0];
+	ret = applet(argc, argv);
+
+	while (argc--)
+		free(argv[argc]);
+	free(argv);
+
+	optind = optind_saved;
+	argv0 = argv0_saved;
+
+	return ret;
+}
+
 static void
 fetch(const char *destdir, const char *src)
 {
@@ -171,6 +243,8 @@ fetch(const char *destdir, const char *src)
 static void
 qmerge_initialize(void)
 {
+	char *buf;
+
 	if (strlen(BUSYBOX))
 		if (access(BUSYBOX, X_OK) != 0)
 			err(BUSYBOX " must be installed");
@@ -186,18 +260,15 @@ qmerge_initialize(void)
 			errp("could not setup PKGDIR: %s", pkgdir);
 	}
 
-	char *buf;
 	xasprintf(&buf, "%s/portage/", port_tmpdir);
 	mkdir_p(buf, 0755);
 	xchdir(buf);
 
-	if (force_download != 2) {
-		if (force_download)
-			unlink(Packages);
-
-		if (access(Packages, R_OK) != 0)
-			fetch(buf, Packages);
+	if (force_download == 1 /* -f: fetch */) {
+		unlink(Packages);
+		fetch(buf, Packages);
 	}
+
 	free(buf);
 }
 
@@ -210,14 +281,23 @@ struct qmerge_bv_state {
 };
 
 static int
-qmerge_filter_cat(q_vdb_cat_ctx *cat_ctx, void *priv)
+qmerge_filter_cat(tree_cat_ctx *cat_ctx, void *priv)
 {
 	struct qmerge_bv_state *state = priv;
 	return !state->catname || strcmp(cat_ctx->name, state->catname) == 0;
 }
 
+/* HACK: pull this in, knowing that qlist will be in the final link, we
+ * should however figure out how to do what match does here from e.g.
+ * atom */
+extern bool qlist_match(
+		tree_pkg_ctx *pkg_ctx,
+		const char *name,
+		depend_atom **name_atom,
+		bool exact);
+
 static int
-qmerge_best_version_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
+qmerge_best_version_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 {
 	struct qmerge_bv_state *state = priv;
 	if (qlist_match(pkg_ctx, state->buf, NULL, true))
@@ -231,6 +311,8 @@ best_version(const char *catname, const char *pkgname, const char *slot)
 {
 	static int vdb_check = 1;
 	static char retbuf[4096];
+
+	tree_ctx *vdb;
 	struct qmerge_bv_state state = {
 		.catname = catname,
 		.pkgname = pkgname,
@@ -257,7 +339,12 @@ best_version(const char *catname, const char *pkgname, const char *slot)
 	retbuf[0] = '\0';
 	snprintf(state.buf, sizeof(state.buf), "%s%s%s:%s",
 		 catname ? : "", catname ? "/" : "", pkgname, slot);
-	q_vdb_foreach_pkg(qmerge_best_version_cb, &state, qmerge_filter_cat);
+	vdb = tree_open_vdb(portroot, portvdb);
+	if (vdb != NULL) {
+		tree_foreach_pkg_fast(vdb,
+				qmerge_best_version_cb, &state, qmerge_filter_cat);
+		tree_close(vdb);
+	}
 
  done:
 	return retbuf;
@@ -268,8 +355,6 @@ config_protected(const char *buf, int cp_argc, char **cp_argv,
                  int cpm_argc, char **cpm_argv)
 {
 	int i;
-	char dest[_Q_PATH_MAX];
-	snprintf(dest, sizeof(dest), "%s%s", portroot, buf);
 
 	/* Check CONFIG_PROTECT_MASK */
 	for (i = 1; i < cpm_argc; ++i)
@@ -279,11 +364,10 @@ config_protected(const char *buf, int cp_argc, char **cp_argv,
 	/* Check CONFIG_PROTECT */
 	for (i = 1; i < cp_argc; ++i)
 		if (strncmp(cp_argv[i], buf, strlen(cp_argv[i])) == 0)
-			if (access(dest, R_OK) == 0)
-				return 1;
+			return 1;
 
 	/* this would probably be bad */
-	if (strcmp("/bin/sh", buf) == 0)
+	if (strcmp(CONFIG_EPREFIX "bin/sh", buf) == 0)
 		return 1;
 
 	return 0;
@@ -538,16 +622,6 @@ install_mask_pwd(int iargc, char **iargv, const struct stat * const st, int fd)
 	install_mask_check_dir(masksv, masksc, st, fd, 1, INCLUDE, qpth);
 }
 
-static char *
-atom2str(const depend_atom *atom, char *buf, size_t size)
-{
-	if (atom->PR_int)
-		snprintf(buf, size, "%s-%s-r%i", atom->PN, atom->PV, atom->PR_int);
-	else
-		snprintf(buf, size, "%s-%s", atom->PN, atom->PV);
-	return buf;
-}
-
 static char
 qprint_tree_node(int level, const depend_atom *atom, const struct pkg_t *pkg)
 {
@@ -570,17 +644,18 @@ qprint_tree_node(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	} else {
 		depend_atom *subatom = atom_explode(p);
 		if (subatom != NULL) {
-			atom2str(subatom, buf, sizeof(buf));
-			atom2str(atom, install_ver, sizeof(install_ver));
-			ret = atom_compare_str(install_ver, buf);
+			ret = atom_compare(atom, subatom);
 			switch (ret) {
 				case EQUAL: c = 'R'; break;
 				case NEWER: c = 'U'; break;
 				case OLDER: c = 'D'; break;
 				default: c = '?'; break;
 			}
-			strncpy(buf, subatom->P, sizeof(buf));
-			snprintf(install_ver, sizeof(install_ver), "[%s%s%s] ", DKBLUE, buf, NORM);
+			snprintf(install_ver, sizeof(install_ver), "[%s%.*s%s] ",
+					DKBLUE,
+					(int)(sizeof(install_ver) - 4 -
+						sizeof(DKBLUE) - sizeof(NORM)),
+					subatom->P, NORM);
 			atom_implode(subatom);
 		}
 		if (update_only && c != 'U')
@@ -660,6 +735,8 @@ pkg_run_func_at(int dirfd, const char *vdb_path, const char *phases, const char
 		"nonfatal() { \"$@\"; }\n"
 		"ebegin() { printf ' * %%b ...' \"$*\"; }\n"
 		"eend() { local r=${1:-$?}; [ $# -gt 0 ] && shift; [ $r -eq 0 ] && echo ' [ ok ]' || echo \" $* \"'[ !! ]'; return $r; }\n"
+		"dodir() { mkdir -p \"$@\"; }\n"
+		"keepdir() { dodir \"$@\" && touch \"$@\"/.keep_${CATEGORY}_${PN}-${SLOT%%/*}; }\n"
 		/* TODO: This should be fatal upon error */
 		"emake() { ${MAKE:-make} ${MAKEOPTS} \"$@\"; }\n"
 		/* Unpack the env if need be */
@@ -670,9 +747,9 @@ pkg_run_func_at(int dirfd, const char *vdb_path, const char *phases, const char
 		"FILESDIR=/.does/not/exist/anywhere\n"
 		"MERGE_TYPE=binary\n"
 		"ROOT='%4$s'\n"
-		"EROOT=\"${EPREFIX%%/}/${ROOT#/}/\"\n"
+		"EROOT=\"/${ROOT#/}/${EPREFIX%%/}/\"\n"
 		"D=\"%5$s\"\n"
-		"ED=\"${EPREFIX%%/}/${D#/}/\"\n"
+		"ED=\"${D%%/}/${EPREFIX%%/}/\"\n"
 		"T=\"%6$s\"\n"
 		/* Finally run the func */
 		"%7$s%2$s\n"
@@ -693,7 +770,7 @@ pkg_run_func_at(int dirfd, const char *vdb_path, const char *phases, const char
 /* Copy one tree (the single package) to another tree (ROOT) */
 static int
 merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
-              FILE *contents, queue **objs, char **cpathp,
+              FILE *contents, size_t eprefix_len, set **objs, char **cpathp,
               int cp_argc, char **cp_argv, int cpm_argc, char **cpm_argv)
 {
 	int i, ret, subfd_src, subfd_dst;
@@ -763,8 +840,9 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 			qprintf("%s>>>%s %s%s%s/\n", GREEN, NORM, DKBLUE, cpath, NORM);
 
 			/* Copy all of these contents */
-			merge_tree_at(subfd_src, name, subfd_dst, name, contents, objs,
-					cpathp, cp_argc, cp_argv, cpm_argc, cpm_argv);
+			merge_tree_at(subfd_src, name,
+					subfd_dst, name, contents, eprefix_len,
+					objs, cpathp, cp_argc, cp_argv, cpm_argc, cpm_argv);
 			cpath = *cpathp;
 			mnlen = 0;
 
@@ -777,29 +855,30 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 			int fd_srcf, fd_dstf;
 			unsigned char *hash;
 			const char *tmpname, *dname;
+			char buf[_Q_PATH_MAX * 2];
+			struct stat ignore;
 
 			/* syntax: obj filename hash mtime */
 			hash = hash_file_at(subfd_src, name, HASH_MD5);
 			if (!pretend)
-				fprintf(contents, "obj %s %s %"PRIu64"\n",
-						cpath, hash, (uint64_t)st.st_mtime);
+				fprintf(contents, "obj %s %s %zu""\n",
+						cpath, hash, (size_t)st.st_mtime);
 			free(hash);
 
 			/* Check CONFIG_PROTECT */
-			if (config_protected(cpath, cp_argc, cp_argv, cpm_argc, cpm_argv)) {
+			if (config_protected(cpath + eprefix_len,
+						cp_argc, cp_argv, cpm_argc, cpm_argv) &&
+					fstatat(subfd_dst, name, &ignore, AT_SYMLINK_NOFOLLOW) == 0)
+			{
 				/* ._cfg####_ */
 				char *num;
-				dname = num = alloca(nlen + 10 + 1);
-				*num++ = '.';
-				*num++ = '_';
-				*num++ = 'c';
-				*num++ = 'f';
-				*num++ = 'g';
-				strcpy(num + 5, name);
+				dname = buf;
+				snprintf(buf, sizeof(buf), "._cfg####_%s", name);
+				num = buf + 5;
 				for (i = 0; i < 10000; ++i) {
 					sprintf(num, "%04i", i);
 					num[4] = '_';
-					if (faccessat(subfd_dst, dname, F_OK, AT_SYMLINK_NOFOLLOW))
+					if (fstatat(subfd_dst, dname, &ignore, AT_SYMLINK_NOFOLLOW))
 						break;
 				}
 				qprintf("%s>>>%s %s (%s)\n", GREEN, NORM, cpath, dname);
@@ -873,19 +952,19 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 		} else if (S_ISLNK(st.st_mode)) {
 			/* Migrate a symlink */
 			size_t len = st.st_size;
-			char *sym = alloca(len + 1);
+			char sym[_Q_PATH_MAX];
 
 			/* Find out what we're pointing to */
-			if (readlinkat(subfd_src, name, sym, len) == -1) {
+			if (readlinkat(subfd_src, name, sym, sizeof(sym)) == -1) {
 				warnp("could not read link %s", cpath);
 				continue;
 			}
-			sym[len] = '\0';
+			sym[len < _Q_PATH_MAX ? len : _Q_PATH_MAX - 1] = '\0';
 
 			/* syntax: sym src -> dst mtime */
 			if (!pretend)
-				fprintf(contents, "sym %s -> %s %"PRIu64"\n",
-						cpath, sym, (uint64_t)st.st_mtime);
+				fprintf(contents, "sym %s -> %s %zu\n",
+						cpath, sym, (size_t)st.st_mtime);
 			qprintf("%s>>>%s %s%s -> %s%s\n", GREEN, NORM,
 					CYAN, cpath, sym, NORM);
 			*objs = add_set(cpath, *objs);
@@ -928,12 +1007,14 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 static void
 pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 {
-	queue *objs;
-	q_vdb_ctx *vdb_ctx;
-	q_vdb_cat_ctx *cat_ctx;
+	set *objs;
+	tree_ctx *vdb;
+	tree_cat_ctx *cat_ctx;
 	FILE *fp, *contents;
 	static char *phases;
 	static size_t phases_len;
+	char *eprefix = NULL;
+	size_t eprefix_len = 0;
 	char buf[1024];
 	char *tbz2, *p, *D, *T;
 	int i;
@@ -1052,19 +1133,19 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	}
 
 	/* Get a handle on the main vdb repo */
-	vdb_ctx = q_vdb_open();
-	if (!vdb_ctx)
+	vdb = tree_open(portroot, portvdb);
+	if (!vdb)
 		return;
-	cat_ctx = q_vdb_open_cat(vdb_ctx, pkg->CATEGORY);
+	cat_ctx = tree_open_cat(vdb, pkg->CATEGORY);
 	if (!cat_ctx) {
 		if (errno != ENOENT) {
-			q_vdb_close(vdb_ctx);
+			tree_close(vdb);
 			return;
 		}
-		mkdirat(vdb_ctx->vdb_fd, pkg->CATEGORY, 0755);
-		cat_ctx = q_vdb_open_cat(vdb_ctx, pkg->CATEGORY);
+		mkdirat(vdb->tree_fd, pkg->CATEGORY, 0755);
+		cat_ctx = tree_open_cat(vdb, pkg->CATEGORY);
 		if (!cat_ctx) {
-			q_vdb_close(vdb_ctx);
+			tree_close(vdb);
 			return;
 		}
 	}
@@ -1156,8 +1237,7 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 				 * needed to decompress the file. See
 				 * https://bugs.gentoo.org/show_bug.cgi?id=634980 */
 				compr = "I zstd --long=31";
-				/*
-				 * If really tar -I would be used we would have to quote:
+				/* If really tar -I would be used we would have to quote:
 				 * compr = "I \"zstd --long=31\"";
 				 * But actually we use a pipe (see below) */
 			} else if (mlen >= 4 &&
@@ -1204,8 +1284,11 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	pkg_run_func("vdb", phases, "pkg_setup", D, T);
 	pkg_run_func("vdb", phases, "pkg_preinst", D, T);
 
+	if (!eat_file("vdb/EPREFIX", &eprefix, &eprefix_len))
+		eprefix_len = 0;
+
 	{
-		int imagefd = open("image" CONFIG_EPREFIX, O_RDONLY);
+		int imagefd = open("image", O_RDONLY);
 		size_t masklen = strlen(install_mask) + 1 +
 				15 + 1 + 14 + 1 + 14 + 1 + 1;  /* worst case scenario */
 		char *imask = xmalloc(masklen);
@@ -1214,6 +1297,12 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 		if (fstat(imagefd, &st) == -1) {
 			close(imagefd);
 			err("Cannot stat image dirfd");
+		} else if (eprefix_len > 0) {
+			int imagepfx = openat(imagefd, eprefix + 1, O_RDONLY);
+			if (imagepfx != -1) {
+				close(imagefd);
+				imagefd = imagepfx;
+			}
 		}
 
 		/* rely on INSTALL_MASK code to remove optional dirs */
@@ -1241,11 +1330,15 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 		close(imagefd);
 	}
 
+	if (eprefix != NULL)
+		free(eprefix);
+
 	makeargv(config_protect, &cp_argc, &cp_argv);
 	makeargv(config_protect_mask, &cpm_argc, &cpm_argv);
 
 	if ((contents = fopen("vdb/CONTENTS", "w")) == NULL)
 		errf("come on wtf?");
+
 	objs = NULL;
 	{
 		char *cpath;
@@ -1253,8 +1346,9 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 
 		cpath = xstrdup("");  /* xrealloced in merge_tree_at */
 
-		ret = merge_tree_at(AT_FDCWD, "image", AT_FDCWD, portroot, contents,
-					&objs, &cpath, cp_argc, cp_argv, cpm_argc, cpm_argv);
+		ret = merge_tree_at(AT_FDCWD, "image",
+				AT_FDCWD, portroot, contents, eprefix_len,
+				&objs, &cpath, cp_argc, cp_argv, cpm_argc, cpm_argv);
 
 		free(cpath);
 
@@ -1275,18 +1369,18 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	/* TODO: Should see about merging with unmerge_packages() */
 	while (1) {
 		int ret;
-		q_vdb_pkg_ctx *pkg_ctx;
+		tree_pkg_ctx *pkg_ctx;
 		depend_atom *old_atom;
 
-		pkg_ctx = q_vdb_next_pkg(cat_ctx);
+		pkg_ctx = tree_next_pkg(cat_ctx);
 		if (!pkg_ctx)
 			break;
-
-		old_atom = atom_explode(pkg_ctx->name);
-		/* This cast sucks, but we know for now the field isn't modified */
-		old_atom->CATEGORY = (char *)cat_ctx->name;
+		old_atom = tree_get_atom(pkg_ctx, 1);  /* retrieve SLOT */
+		if (!old_atom)
+			goto next_pkg;
+		old_atom->SUBSLOT = NULL;  /* just match SLOT */
+		old_atom->REPO = NULL;     /* REPO never matters, TODO atom_compare */
 		ret = atom_compare(atom, old_atom);
-		atom_implode(old_atom);
 		switch (ret) {
 			case NEWER:
 			case OLDER:
@@ -1301,24 +1395,25 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 				goto next_pkg;
 		}
 
-		qprintf("%s+++%s %s/%s %s %s/%s\n",
+		printf("%s+++%s %s/%s %s %s/%s\n",
 				GREEN, NORM, atom->CATEGORY, pkg->PF,
 				booga[ret], cat_ctx->name, pkg_ctx->name);
 
 		pkg_unmerge(pkg_ctx, objs, cp_argc, cp_argv, cpm_argc, cpm_argv);
  next_pkg:
-		q_vdb_close_pkg(pkg_ctx);
+		tree_close_pkg(pkg_ctx);
 	}
 
 	freeargv(cp_argc, cp_argv);
 	freeargv(cpm_argc, cpm_argv);
 
 	/* Clean up the package state */
-	free_sets(objs);
+	free_set(objs);
 	free(D);
 	free(T);
 
 	/* Update the magic counter */
+	/* FIXME: check Portage's get_counter_tick_core */
 	if ((fp = fopen("vdb/COUNTER", "w")) != NULL) {
 		fputs("0", fp);
 		fclose(fp);
@@ -1330,11 +1425,9 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 				portroot, portvdb, pkg->CATEGORY);
 		mkdir_p(buf, 0755);
 		strcat(buf, pkg->PF);
-		if (rename("vdb", buf)) {
-			xasprintf(&p, "mv vdb '%s'", buf);
-			xsystem(p);
-			free(p);
-		}
+		rm_rf(buf);  /* get rid of existing dir, empty dir is fine */
+		if (rename("vdb", buf) != 0)
+			warn("failed to move 'vdb' to '%s': %s", buf, strerror(errno));
 	}
 
 	/* clean up our local temp dir */
@@ -1344,21 +1437,21 @@ pkg_merge(int level, const depend_atom *atom, const struct pkg_t *pkg)
 	rmdir("../qmerge");
 
 	printf("%s>>>%s %s%s%s/%s%s%s\n",
-			YELLOW, NORM, WHITE, atom->CATEGORY, NORM, CYAN, atom->PN, NORM);
+			YELLOW, NORM, WHITE, atom->CATEGORY, NORM, CYAN, pkg->PF, NORM);
 
-	q_vdb_close(vdb_ctx);
+	tree_close(vdb);
 }
 
 static int
-pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
+pkg_unmerge(tree_pkg_ctx *pkg_ctx, set *keep,
 		int cp_argc, char **cp_argv, int cpm_argc, char **cpm_argv)
 {
-	q_vdb_cat_ctx *cat_ctx = pkg_ctx->cat_ctx;
-	const char *cat = cat_ctx->name;
-	const char *pkgname = pkg_ctx->name;
+	tree_cat_ctx *cat_ctx = pkg_ctx->cat_ctx;
 	size_t buflen;
 	static char *phases;
 	static size_t phases_len;
+	char *eprefix = NULL;
+	size_t eprefix_len = 0;
 	const char *T;
 	char *buf;
 	FILE *fp;
@@ -1370,14 +1463,14 @@ pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
 	buf = phases = NULL;
 	T = "${PWD}/temp";
 
-	printf("%s<<<%s %s%s%s/%s%s%s\n",
-			YELLOW, NORM, WHITE, cat, NORM, CYAN, pkgname, NORM);
+	printf("%s<<<%s %s\n", YELLOW, NORM,
+			atom_format("%[CATEGORY]%[PF]", tree_get_atom(pkg_ctx, false)));
 
 	if (pretend == 100)
 		return 0;
 
 	/* First get a handle on the things to clean up */
-	fp = q_vdb_pkg_fopenat_ro(pkg_ctx, "CONTENTS");
+	fp = tree_pkg_vdb_fopenat_ro(pkg_ctx, "CONTENTS");
 	if (fp == NULL)
 		return ret;
 
@@ -1385,16 +1478,21 @@ pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
 
 	/* Then execute the pkg_prerm step */
 	if (!pretend) {
-		q_vdb_pkg_eat(pkg_ctx, "DEFINED_PHASES", &phases, &phases_len);
+		tree_pkg_vdb_eat(pkg_ctx, "DEFINED_PHASES", &phases, &phases_len);
 		mkdirat(pkg_ctx->fd, "temp", 0755);
 		pkg_run_func_at(pkg_ctx->fd, ".", phases, "pkg_prerm", T, T);
 	}
 
+	if (!tree_pkg_vdb_eat(pkg_ctx, "EPREFIX", &eprefix, &eprefix_len))
+		eprefix_len = 0;
+	if (eprefix != NULL)
+		free(eprefix);
+
 	unmerge_config_protected =
 		strstr(features, "config-protect-if-modified") != NULL;
 
 	while (getline(&buf, &buflen, fp) != -1) {
-		queue *q;
+		bool del;
 		contents_entry *e;
 		char zing[20];
 		int protected = 0;
@@ -1404,7 +1502,7 @@ pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
 		if (!e)
 			continue;
 
-		protected = config_protected(e->name,
+		protected = config_protected(e->name + eprefix_len,
 				cp_argc, cp_argv, cpm_argc, cpm_argv);
 
 		/* This should never happen ... */
@@ -1464,21 +1562,16 @@ pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
 		}
 
 		/* See if this was updated */
-		q = keep;
-		while (q) {
-			if (!strcmp(q->name, e->name)) {
-				/* XXX: could remove this from the queue */
-				strcpy(zing, "---");
-				q = NULL;
-				break;
-			}
-			q = q->next;
-		}
+		del = false;
+		if (keep != NULL)
+			(void)del_set(e->name, keep, &del);
+		if (del)
+			strcpy(zing, "---");
 
 		/* No match, so unmerge it */
 		if (!quiet)
 			printf("%s %s\n", zing, e->name);
-		if (!keep || q) {
+		if (!keep || !del) {
 			char *p;
 
 			if (!pretend && unlinkat(portroot_fd, e->name + 1, 0)) {
@@ -1522,7 +1615,7 @@ pkg_unmerge(q_vdb_pkg_ctx *pkg_ctx, queue *keep,
 		unlinkat(cat_ctx->fd, pkg_ctx->name, AT_REMOVEDIR);
 
 		/* And prune the category if it's empty */
-		unlinkat(cat_ctx->ctx->vdb_fd, cat_ctx->name, AT_REMOVEDIR);
+		unlinkat(cat_ctx->ctx->tree_fd, cat_ctx->name, AT_REMOVEDIR);
 	}
 
 	ret = 0;
@@ -1608,7 +1701,9 @@ pkg_fetch(int level, const depend_atom *atom, const struct pkg_t *pkg)
 
 	/* XXX: should do a size check here for partial downloads */
 
-	if (force_download && (access(buf, R_OK) == 0) && (pkg->SHA1[0] || pkg->MD5[0])) {
+	if (force_download && (access(buf, R_OK) == 0) &&
+			(pkg->SHA1[0] || pkg->MD5[0]))
+	{
 		if (pkg_verify_checksums(buf, pkg, atom, 0, 0) != 0)
 			unlink(buf);
 	}
@@ -1617,7 +1712,9 @@ pkg_fetch(int level, const depend_atom *atom, const struct pkg_t *pkg)
 			warn("No checksum data for %s (try `emaint binhost --fix`)", buf);
 			return;
 		} else {
-			if (pkg_verify_checksums(buf, pkg, atom, qmerge_strict, !quiet) == 0) {
+			if (pkg_verify_checksums(buf, pkg, atom, qmerge_strict, !quiet)
+					== 0)
+			{
 				pkg_merge(0, atom, pkg);
 				return;
 			}
@@ -1662,7 +1759,7 @@ print_Pkg(int full, const depend_atom *atom, const struct pkg_t *pkg)
 	char *p = NULL;
 	char buf[512];
 
-	printf("%s%s/%s%s%s%s%s%s\n", BOLD, atom->CATEGORY, BLUE, pkg->PF, NORM,
+	printf("%s%s/%s%s:%s%s%s%s%s\n", BOLD, atom->CATEGORY, BLUE, pkg->PF, pkg->SLOT, NORM,
 		!quiet ? " [" : "",
 		!quiet ? make_human_readable_str(pkg->SIZE, 1, KILOBYTE) : "",
 		!quiet ? " KiB]" : "");
@@ -1707,23 +1804,25 @@ print_Pkg(int full, const depend_atom *atom, const struct pkg_t *pkg)
 }
 
 static int
-qmerge_unmerge_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
+qmerge_unmerge_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 {
-	queue *todo = priv;
 	int cp_argc;
 	int cpm_argc;
 	char **cp_argv;
 	char **cpm_argv;
+	char **todo;
+	char **p;
 
 	makeargv(config_protect, &cp_argc, &cp_argv);
 	makeargv(config_protect_mask, &cpm_argc, &cpm_argv);
 
-	while (todo) {
-		if (qlist_match(pkg_ctx, todo->name, NULL, true))
+	(void)list_set(priv, &todo);
+	for (p = todo; *p != NULL; p++) {
+		if (qlist_match(pkg_ctx, *p, NULL, true))
 			pkg_unmerge(pkg_ctx, NULL, cp_argc, cp_argv, cpm_argc, cpm_argv);
-		todo = todo->next;
 	}
 
+	free(todo);
 	freeargv(cp_argc, cp_argv);
 	freeargv(cpm_argc, cpm_argv);
 
@@ -1731,9 +1830,15 @@ qmerge_unmerge_cb(q_vdb_pkg_ctx *pkg_ctx, void *priv)
 }
 
 static int
-unmerge_packages(queue *todo)
+unmerge_packages(set *todo)
 {
-	return q_vdb_foreach_pkg(qmerge_unmerge_cb, todo, NULL);
+	tree_ctx *vdb = tree_open_vdb(portroot, portvdb);
+	int ret = 1;
+	if (vdb != NULL) {
+		ret = tree_foreach_pkg_fast(vdb, qmerge_unmerge_cb, todo, NULL);
+		tree_close(vdb);
+	}
+	return ret;
 }
 
 static FILE *
@@ -1814,23 +1919,38 @@ grab_binpkg_info(const char *name)
 				snprintf(buf, sizeof(buf), "%s/%s", pkg->CATEGORY, pkg->PF);
 				if (strstr(buf, name) != NULL) {
 					if (!best_match[0])
-						strncpy(best_match, buf, sizeof(best_match));
+						snprintf(best_match, sizeof(best_match), "%.*s",
+								(int)sizeof(best_match) - 1, buf);
 
 					atom = atom_explode(buf);
-					snprintf(buf, sizeof(buf), "%s/%s-%s", atom->CATEGORY, atom->PN, atom->PV);
-					if (atom->PR_int)
-						snprintf(buf, sizeof(buf), "%s/%s-%s-r%i", atom->CATEGORY, atom->PN, atom->PV, atom->PR_int);
+					if (atom->PR_int) {
+						snprintf(buf, sizeof(buf), "%s/%s-%s-r%i",
+								atom->CATEGORY, atom->PN,
+								atom->PV, atom->PR_int);
+					} else {
+						snprintf(buf, sizeof(buf), "%s/%s-%s",
+								atom->CATEGORY, atom->PN, atom->PV);
+					}
 					ret = atom_compare_str(name, buf);
-					IF_DEBUG(fprintf(stderr, "=== atom_compare(%s, %s) = %d %s\n", name, buf, ret, booga[ret])); /* buf(%s) depend(%s)\n", ret, pkg->CATEGORY, pkg->PF, name, pkg->RDEPEND); */
+					IF_DEBUG(fprintf(stderr,
+								"=== atom_compare(%s, %s) = %d %s\n",
+								name, buf, ret, booga[ret]));
+					/* buf(%s) depend(%s)\n", ret, pkg->CATEGORY,
+					 * pkg->PF, name, pkg->RDEPEND); */
 					switch (ret) {
 						case EQUAL:
 						case NEWER:
-							snprintf(buf, sizeof(buf), "%s/%s", pkg->CATEGORY, pkg->PF);
+							snprintf(buf, sizeof(buf), "%s/%s",
+									pkg->CATEGORY, pkg->PF);
 							ret = atom_compare_str(buf, best_match);
 							if (ret == NEWER || ret == EQUAL) {
-								strncpy(best_match, buf, sizeof(best_match));
+								snprintf(best_match, sizeof(best_match), "%.*s",
+										(int)sizeof(best_match) - 1, buf);
 								memcpy(rpkg, pkg, sizeof(struct pkg_t));
-								IF_DEBUG(fprintf(stderr, "--- %s/%s depend(%s)\n", rpkg->CATEGORY, rpkg->PF, rpkg->RDEPEND));
+								IF_DEBUG(fprintf(stderr,
+											"--- %s/%s depend(%s)\n",
+											rpkg->CATEGORY, rpkg->PF,
+											rpkg->RDEPEND));
 							}
 						case OLDER: break;
 						default:
@@ -1856,34 +1976,46 @@ grab_binpkg_info(const char *name)
 		if (*buf) {
 			/* we dont need all the info */
 			if (strcmp(buf, "RDEPEND") == 0)
-				strncpy(pkg->RDEPEND, p, sizeof(Pkg.RDEPEND));
+				snprintf(pkg->RDEPEND, sizeof(Pkg.RDEPEND), "%.*s",
+						(int)sizeof(Pkg.RDEPEND) - 1, p);
 			if (strcmp(buf, "PF") == 0)
-				strncpy(pkg->PF, p, sizeof(Pkg.PF));
+				snprintf(pkg->PF, sizeof(Pkg.PF), "%.*s",
+						(int)sizeof(Pkg.PF) - 1, p);
 			if (strcmp(buf, "CATEGORY") == 0)
-				strncpy(pkg->CATEGORY, p, sizeof(Pkg.CATEGORY));
+				snprintf(pkg->CATEGORY, sizeof(Pkg.CATEGORY), "%.*s",
+						(int)sizeof(Pkg.CATEGORY) - 1, p);
 			if (strcmp(buf, "REPO") == 0)
-				strncpy(pkg->REPO, p, sizeof(Pkg.REPO));
+				snprintf(pkg->REPO, sizeof(Pkg.REPO), "%.*s",
+						(int)sizeof(Pkg.REPO) - 1, p);
 
 			if (strcmp(buf, "CPV") == 0) {
 				if ((atom = atom_explode(p)) != NULL) {
-					snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
-					if (atom->PR_int)
-						snprintf(buf, sizeof(buf), "%s-%s-r%i", atom->PN, atom->PV, atom->PR_int);
-					strncpy(pkg->PF, buf, sizeof(Pkg.PF));
-					strncpy(pkg->CATEGORY, atom->CATEGORY, sizeof(Pkg.CATEGORY));
+					if (atom->PR_int) {
+						snprintf(buf, sizeof(buf), "%s-%s-r%i",
+								atom->PN, atom->PV, atom->PR_int);
+					} else {
+						snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
+					}
+					snprintf(pkg->PF, sizeof(Pkg.PF), "%.*s",
+							(int)sizeof(Pkg.PF) - 1, buf);
+					snprintf(pkg->CATEGORY, sizeof(Pkg.CATEGORY), "%.*s",
+							(int)sizeof(Pkg.CATEGORY) - 1, atom->CATEGORY);
 					atom_implode(atom);
 				}
 			}
 			if (strcmp(buf, "SLOT") == 0)
-				strncpy(pkg->SLOT, p, sizeof(Pkg.SLOT));
+				snprintf(pkg->SLOT, sizeof(Pkg.SLOT), "%.*s",
+						(int)sizeof(Pkg.SLOT) - 1, p);
 			if (strcmp(buf, "USE") == 0)
-				strncpy(pkg->USE, p, sizeof(Pkg.USE));
-
+				snprintf(pkg->USE, sizeof(Pkg.USE), "%.*s",
+						(int)sizeof(Pkg.USE) - 1, p);
 			/* checksums. We must have 1 or the other unless --*/
 			if (strcmp(buf, "MD5") == 0)
-				strncpy(pkg->MD5, p, sizeof(Pkg.MD5));
+				snprintf(pkg->MD5, sizeof(Pkg.MD5), "%.*s",
+						(int)sizeof(Pkg.MD5) - 1, p);
 			if (strcmp(buf, "SHA1") == 0)
-				strncpy(pkg->SHA1, p, sizeof(Pkg.SHA1));
+				snprintf(pkg->SHA1, sizeof(Pkg.SHA1), "%.*s",
+						(int)sizeof(Pkg.SHA1) - 1, p);
 		}
 	}
 	fclose(fp);
@@ -1918,10 +2050,12 @@ find_binpkg(const char *name)
 					depend_atom *atom;
 
 					if (!best_match[0])
-						strncpy(best_match, buf, sizeof(best_match));
+						snprintf(best_match, sizeof(best_match), "%.*s",
+								(int)sizeof(best_match) - 1, buf);
 
 					atom = atom_explode(buf);
-					snprintf(buf, sizeof(buf), "%s/%s", atom->CATEGORY, atom->PN);
+					snprintf(buf, sizeof(buf), "%s/%s",
+							atom->CATEGORY, atom->PN);
 					ret = atom_compare_str(name, buf);
 					switch (ret) {
 						case OLDER: break;
@@ -1930,8 +2064,10 @@ find_binpkg(const char *name)
 							snprintf(buf, sizeof(buf), "%s/%s", CATEGORY, PF);
 							ret = atom_compare_str(buf, best_match);
 							if (ret == NEWER || ret == EQUAL)
-								strncpy(best_match, buf, sizeof(best_match));
-							/* printf("[%s == %s] = %d; %s/%s\n", name, buf, ret, CATEGORY, PF); */
+								snprintf(best_match, sizeof(best_match), "%.*s",
+										(int)sizeof(best_match) - 1, buf);
+							/* printf("[%s == %s] = %d; %s/%s\n",
+							 * name, buf, ret, CATEGORY, PF); */
 						default:
 							break;
 					}
@@ -1954,18 +2090,23 @@ find_binpkg(const char *name)
 			if (strcmp(buf, "CPV") == 0) {
 				depend_atom *atom;
 				if ((atom = atom_explode(p)) != NULL) {
-					snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
-					if (atom->PR_int)
-						snprintf(buf, sizeof(buf), "%s-%s-r%i", atom->PN, atom->PV, atom->PR_int);
-					strncpy(PF, buf, sizeof(PF));
-					strncpy(CATEGORY, atom->CATEGORY, sizeof(CATEGORY));
+					if (atom->PR_int) {
+						snprintf(buf, sizeof(buf), "%s-%s-r%i",
+								atom->PN, atom->PV, atom->PR_int);
+					} else {
+						snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PV);
+					}
+					snprintf(PF, sizeof(PF), "%.*s", (int)sizeof(PF) - 1, buf);
+					snprintf(CATEGORY, sizeof(CATEGORY), "%.*s",
+							(int)sizeof(CATEGORY) - 1, atom->CATEGORY);
 					atom_implode(atom);
 				}
 			}
 			if (strcmp(buf, "PF") == 0)
-				strncpy(PF, p, sizeof(PF));
+				snprintf(PF, sizeof(PF), "%.*s", (int)sizeof(PF) - 1, p);
 			if (strcmp(buf, "CATEGORY") == 0)
-				strncpy(CATEGORY, p, sizeof(CATEGORY));
+				snprintf(CATEGORY, sizeof(CATEGORY), "%.*s",
+						(int)sizeof(CATEGORY) - 1, p);
 		}
 	}
 	fclose(fp);
@@ -1973,7 +2114,7 @@ find_binpkg(const char *name)
 }
 
 static int
-parse_packages(queue *todo)
+parse_packages(set *todo)
 {
 	FILE *fp;
 	int linelen;
@@ -1982,10 +2123,16 @@ parse_packages(queue *todo)
 	struct pkg_t Pkg;
 	depend_atom *pkg_atom;
 	char repo[sizeof(Pkg.REPO)];
+	depend_atom **todo_atoms;
+	size_t todo_cnt;
+	size_t i;
 
 	fp = open_binpkg_index();
+	if (fp == NULL)
+		return EXIT_FAILURE;
 
 	buf = NULL;
+	buflen = 0;  /* make getline allocate */
 	repo[0] = '\0';
 
 	/* First consume the header with the common data. */
@@ -2004,7 +2151,7 @@ parse_packages(queue *todo)
 		switch (*buf) {
 		case 'R':
 			if (!strcmp(buf, "REPO"))
-				strncpy(repo, p, sizeof(repo));
+				snprintf(repo, sizeof(repo), "%.*s", (int)sizeof(repo) - 1, p);
 			break;
 		}
 	}
@@ -2013,6 +2160,16 @@ parse_packages(queue *todo)
 	memset(&Pkg, 0, sizeof(Pkg));
 	strcpy(Pkg.SLOT, "0");
 
+	/* build list with exploded atoms for each access below */
+	{
+		char **todo_strs;
+		todo_cnt = list_set(todo, &todo_strs);
+		todo_atoms = xmalloc(sizeof(*todo_atoms) * todo_cnt);
+		for (i = 0; i < todo_cnt; i++)
+			todo_atoms[i] = atom_explode(todo_strs[i]);
+		free(todo_strs);
+	}
+
 	/* Then walk all the package entries. */
 	while (getline(&buf, &buflen, fp) != -1) {
 		if (*buf == '\n') {
@@ -2020,20 +2177,15 @@ parse_packages(queue *todo)
 				if (search_pkgs && !todo) {
 					print_Pkg(verbose, pkg_atom, &Pkg);
 				} else {
-					const queue *ll = todo;
-					depend_atom *todo_atom;
-					while (ll) {
-						todo_atom = atom_explode(ll->name);
-						pkg_atom->REPO = todo_atom->REPO ? Pkg.REPO : NULL;
-						pkg_atom->SLOT = todo_atom->SLOT ? Pkg.SLOT : NULL;
-						if (atom_compare(todo_atom, pkg_atom) == EQUAL) {
+					for (i = 0; i < todo_cnt; i++) {
+						pkg_atom->REPO = todo_atoms[i]->REPO ? Pkg.REPO : NULL;
+						pkg_atom->SLOT = todo_atoms[i]->SLOT ? Pkg.SLOT : NULL;
+						if (atom_compare(pkg_atom, todo_atoms[i]) == EQUAL) {
 							if (search_pkgs)
 								print_Pkg(verbose, pkg_atom, &Pkg);
 							else
 								pkg_fetch(0, pkg_atom, &Pkg);
 						}
-						atom_implode(todo_atom);
-						ll = ll->next;
 					}
 				}
 
@@ -2057,40 +2209,66 @@ parse_packages(queue *todo)
 
 		switch (*buf) {
 			case 'U':
-				if (strcmp(buf, "USE") == 0) strncpy(Pkg.USE, p, sizeof(Pkg.USE));
+				if (strcmp(buf, "USE") == 0)
+					snprintf(Pkg.USE, sizeof(Pkg.USE), "%.*s",
+							(int)sizeof(Pkg.USE) - 1, p);
 				break;
 			case 'P':
-				if (strcmp(buf, "PF") == 0) strncpy(Pkg.PF, p, sizeof(Pkg.PF));
+				if (strcmp(buf, "PF") == 0)
+					snprintf(Pkg.PF, sizeof(Pkg.PF), "%.*s",
+							(int)sizeof(Pkg.PF) - 1, p);
 				break;
 			case 'S':
-				if (strcmp(buf, "SIZE") == 0) Pkg.SIZE = atol(p);
-				if (strcmp(buf, "SLOT") == 0) strncpy(Pkg.SLOT, p, sizeof(Pkg.SLOT));
-				if (strcmp(buf, "SHA1") == 0) strncpy(Pkg.SHA1, p, sizeof(Pkg.SHA1));
+				if (strcmp(buf, "SIZE") == 0)
+					Pkg.SIZE = atol(p);
+				if (strcmp(buf, "SLOT") == 0)
+					snprintf(Pkg.SLOT, sizeof(Pkg.SLOT), "%.*s",
+							(int)sizeof(Pkg.SLOT) - 1, p);
+				if (strcmp(buf, "SHA1") == 0)
+					snprintf(Pkg.SHA1, sizeof(Pkg.SHA1), "%.*s",
+							(int)sizeof(Pkg.SHA1) - 1, p);
 				break;
 			case 'M':
-				if (strcmp(buf, "MD5") == 0) strncpy(Pkg.MD5, p, sizeof(Pkg.MD5));
+				if (strcmp(buf, "MD5") == 0)
+					snprintf(Pkg.MD5, sizeof(Pkg.MD5), "%.*s",
+							(int)sizeof(Pkg.MD5) - 1, p);
 				break;
 			case 'R':
-				if (strcmp(buf, "REPO") == 0) strncpy(Pkg.REPO, p, sizeof(Pkg.REPO));
-				if (strcmp(buf, "RDEPEND") == 0) strncpy(Pkg.RDEPEND, p, sizeof(Pkg.RDEPEND));
+				if (strcmp(buf, "REPO") == 0)
+					snprintf(Pkg.REPO, sizeof(Pkg.REPO), "%.*s",
+							(int)sizeof(Pkg.REPO) - 1, p);
+				if (strcmp(buf, "RDEPEND") == 0)
+					snprintf(Pkg.RDEPEND, sizeof(Pkg.RDEPEND), "%.*s",
+							(int)sizeof(Pkg.RDEPEND) - 1, p);
 				break;
 			case 'L':
-				if (strcmp(buf, "LICENSE") == 0) strncpy(Pkg.LICENSE, p, sizeof(Pkg.LICENSE));
+				if (strcmp(buf, "LICENSE") == 0)
+					snprintf(Pkg.LICENSE, sizeof(Pkg.LICENSE), "%.*s",
+							(int)sizeof(Pkg.LICENSE) - 1, p);
 				break;
 			case 'C':
-				if (strcmp(buf, "CATEGORY") == 0) strncpy(Pkg.CATEGORY, p, sizeof(Pkg.CATEGORY));
+				if (strcmp(buf, "CATEGORY") == 0)
+					snprintf(Pkg.CATEGORY, sizeof(Pkg.CATEGORY), "%.*s",
+							(int)sizeof(Pkg.CATEGORY) - 1, p);
 				if (strcmp(buf, "CPV") == 0) {
 					if ((pkg_atom = atom_explode(p)) != NULL) {
 						if (pkg_atom->PR_int)
-							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s-r%i", pkg_atom->PN, pkg_atom->PV, pkg_atom->PR_int);
+							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s-r%i",
+									pkg_atom->PN, pkg_atom->PV,
+									pkg_atom->PR_int);
 						else
-							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s", pkg_atom->PN, pkg_atom->PV);
-						strncpy(Pkg.CATEGORY, pkg_atom->CATEGORY, sizeof(Pkg.CATEGORY));
+							snprintf(Pkg.PF, sizeof(Pkg.PF), "%s-%s",
+									pkg_atom->PN, pkg_atom->PV);
+						snprintf(Pkg.CATEGORY, sizeof(Pkg.CATEGORY),
+								"%.*s", (int)sizeof(Pkg.CATEGORY) - 1,
+								pkg_atom->CATEGORY);
 					}
 				}
 				break;
 			case 'D':
-				if (strcmp(buf, "DESC") == 0) strncpy(Pkg.DESC, p, sizeof(Pkg.DESC));
+				if (strcmp(buf, "DESC") == 0)
+					snprintf(Pkg.DESC, sizeof(Pkg.DESC), "%.*s",
+							(int)sizeof(Pkg.DESC) - 1, p);
 				break;
 			default:
 				break;
@@ -2102,11 +2280,15 @@ parse_packages(queue *todo)
 	if (pkg_atom)
 		atom_implode(pkg_atom);
 
+	for (i = 0; i < todo_cnt; i++)
+		atom_implode(todo_atoms[i]);
+	free(todo_atoms);
+
 	return EXIT_SUCCESS;
 }
 
-static queue *
-qmerge_add_set_file(const char *dir, const char *file, queue *set)
+static set *
+qmerge_add_set_file(const char *pfx, const char *dir, const char *file, set *q)
 {
 	FILE *fp;
 	int linelen;
@@ -2114,7 +2296,7 @@ qmerge_add_set_file(const char *dir, const char *file, queue *set)
 	char *buf, *fname;
 
 	/* Find the file to read */
-	xasprintf(&fname, "%s%s/%s", portroot, dir, file);
+	xasprintf(&fname, "%s%s%s/%s", portroot, pfx, dir, file);
 
 	if ((fp = fopen(fname, "r")) == NULL) {
 		warnp("unable to read set file %s", fname);
@@ -2127,19 +2309,19 @@ qmerge_add_set_file(const char *dir, const char *file, queue *set)
 	buf = NULL;
 	while ((linelen = getline(&buf, &buflen, fp)) >= 0) {
 		rmspace_len(buf, (size_t)linelen);
-		set = add_set(buf, set);
+		q = add_set(buf, q);
 	}
 	free(buf);
 
 	fclose(fp);
 
-	return set;
+	return q;
 }
 
 static void *
 qmerge_add_set_system(void *data, char *buf)
 {
-	queue *set = data;
+	set *q = data;
 	char *s;
 
 	s = strchr(buf, '#');
@@ -2149,35 +2331,45 @@ qmerge_add_set_system(void *data, char *buf)
 
 	s = buf;
 	if (*s == '*')
-		set = add_set(s + 1, set);
+		q = add_set(s + 1, q);
 	else if (s[0] == '-' && s[1] == '*') {
-		int ok;
-		set = del_set(s + 2, set, &ok);
+		bool ok;
+		q = del_set(s + 2, q, &ok);
 	}
 
-	return set;
+	return q;
 }
 
 /* XXX: note, this doesn't handle more complicated set files like
  *      the portage .ini files in /usr/share/portage/sets/ */
 /* XXX: this code does not combine duplicate dependencies */
-static queue *
-qmerge_add_set(char *buf, queue *set)
+static set *
+qmerge_add_set(char *buf, set *q)
 {
 	if (strcmp(buf, "world") == 0)
-		return qmerge_add_set_file("/var/lib/portage", "world", set);
-	else if (strcmp(buf, "all") == 0)
-		return get_vdb_atoms(0);
-	else if (strcmp(buf, "system") == 0)
-		return q_profile_walk("packages", qmerge_add_set_system, set);
+		return qmerge_add_set_file(CONFIG_EPREFIX, "/var/lib/portage",
+				"world", q);
+	else if (strcmp(buf, "all") == 0) {
+		tree_ctx *ctx = tree_open_vdb(portroot, portvdb);
+		set *ret = NULL;
+		if (ctx != NULL) {
+			ret = tree_get_atoms(ctx, false, NULL);
+			tree_close(ctx);
+		}
+		return ret;
+	} else if (strcmp(buf, "system") == 0)
+		return q_profile_walk("packages", qmerge_add_set_system, q);
 	else if (buf[0] == '@')
-		return qmerge_add_set_file("/etc/portage", buf+1, set);
-	else
-		return add_set(buf, set);
+		/* TODO: use configroot */
+		return qmerge_add_set_file(CONFIG_EPREFIX, "/etc/portage", buf+1, q);
+	else {
+		rmspace(buf);
+		return add_set(buf, q);
+	}
 }
 
 static int
-qmerge_run(queue *todo)
+qmerge_run(set *todo)
 {
 	if (uninstall)
 		return unmerge_packages(todo);
@@ -2188,7 +2380,7 @@ qmerge_run(queue *todo)
 int qmerge_main(int argc, char **argv)
 {
 	int i, ret;
-	queue *todo;
+	set *todo;
 
 	if (argc < 2)
 		qmerge_usage(EXIT_FAILURE);
@@ -2227,6 +2419,11 @@ int qmerge_main(int argc, char **argv)
 	for (i = optind; i < argc; ++i)
 		todo = qmerge_add_set(argv[i], todo);
 
+	if (todo == NULL) {
+		warn("need package names to work with");
+		return EXIT_FAILURE;
+	}
+
 	if (!uninstall)
 		qmerge_initialize();
 
@@ -2257,10 +2454,6 @@ int qmerge_main(int argc, char **argv)
 	}
 
 	ret = qmerge_run(todo);
-	free_sets(todo);
+	free_set(todo);
 	return ret;
 }
-
-#else
-DEFINE_APPLET_STUB(qmerge)
-#endif
